use crate::span::Span;

/// Errors that can occur during lexical analysis.
#[derive(Debug, Clone, PartialEq, Eq)]
pub enum LexerErrorKind {
    /// An invalid character was encountered.
    InvalidChar,
    /// An invalid escape sequence was encountered.
    InvalidEscape(char),
    /// An invalid number was encountered.
    InvalidNumber(String),
    /// An invalid string was encountered.
    InvalidString,
    /// An unclosed character literal was encountered.
    UnclosedChar,
    /// And unclosed string literal was encountered.
    UnclosedString,
}

/// Lexer error, with a start and end location.
#[derive(Debug, Clone, PartialEq)]
pub struct LexerError {
    /// The type of error encountered.
    pub kind: LexerErrorKind,
    /// The span in which the error occurred.
    pub span: Span,
}

impl LexerError {
    /// Construct a new instance of `LexerError`.
    #[must_use]
    pub const fn new(kind: LexerErrorKind, span: Span) -> Self {
        Self { kind, span }
    }
}

impl std::error::Error for LexerError {}

impl std::fmt::Display for LexerError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        use LexerErrorKind::*;

        match &self.kind {
            InvalidChar => write!(f, "Invalid character literal"),
            InvalidEscape(c) => write!(f, "Unknown escape sequence '\\{c}' in string"),
            InvalidNumber(n) => write!(f, "`{n}` is not a valid numeric literal"),
            InvalidString => write!(f, "Invalid string literal"),
            UnclosedChar => write!(f, "Unclosed character literal"),
            UnclosedString => write!(f, "Unclosed string literal"),
        }
    }
}
