use std::{
    str::{Chars, FromStr},
    sync::Arc,
};

pub use self::{
    error::{LexerError, LexerErrorKind},
    symbol::Symbol,
    token::{Token, TokenKind},
};
use crate::span::{Source, Span};

mod error;
mod symbol;
mod token;

/// A trait for checking if a character is a separator.
pub trait Separator {
    /// Check if the character is a separator.
    fn is_separator(&self) -> bool;
}

impl Separator for char {
    fn is_separator(&self) -> bool {
        self.is_ascii_whitespace() || matches!(self, '(' | ')' | '[' | ']' | '{' | '}')
    }
}

/// A lexer, used by the parser.
#[derive(Debug)]
pub struct Lexer<'lexer> {
    input: Chars<'lexer>,
    byte: usize,
    source: Arc<Source>,
}

impl<'lexer> Lexer<'lexer> {
    /// Create a new lexer instance from a string.
    #[must_use]
    pub fn new(input: &'lexer str) -> Self {
        let source = Arc::new(Source::new(None, input.to_string()));

        Self {
            input: input.chars(),
            byte: 0,
            source,
        }
    }

    /// Set the name of the [Source] being lexically analyzed.
    pub fn set_name(&mut self, name: String) {
        // FIXME: Avoid unwrapping here (if possible?)
        Arc::get_mut(&mut self.source).unwrap().set_name(name);
    }

    /// The source being lexically analyzed.
    #[must_use]
    pub fn source(&self) -> Arc<Source> {
        self.source.clone()
    }

    /// Get the unparsed input.
    #[must_use]
    pub fn get_unparsed(&self) -> &str {
        self.input.as_str()
    }

    /// Get the current position of the lexer.
    #[inline]
    #[must_use]
    pub(crate) fn span(&self) -> Span {
        Span::new(self.byte..self.byte, self.source.clone())
    }

    /// Get the current character.
    #[must_use]
    fn current(&self) -> Option<char> {
        self.input.as_str().chars().next()
    }

    /// Get the nth character ahead of the current character without advancing.
    #[inline]
    fn peek(&self, n: usize) -> Option<char> {
        self.input.as_str().chars().nth(n)
    }

    /// Advance the lexer by one character.
    fn advance(&mut self) -> Option<char> {
        let c = self.input.next()?;
        self.byte += c.len_utf8();

        Some(c)
    }

    /// Read a word from the input until a separator is reached.
    fn read_word(&mut self) -> String {
        let mut word = String::new();
        while let Some(c) = self.current() {
            if c.is_separator() || (c == '#' && self.peek(1) == Some('{')) {
                break;
            }

            word.push(c);
            self.advance();
        }

        word
    }

    /// Parse a value from the input or return an error.
    fn parse_or<T>(&mut self, err: impl Fn(String) -> LexerErrorKind) -> Result<T, LexerError>
    where
        T: FromStr,
    {
        let span = self.span();
        let word = self.read_word();

        word.parse()
            .map_err(|_| LexerError::new(err(word), span.join(&self.span())))
    }

    /// Read the next token from the input.
    pub fn read(&mut self) -> Result<Option<Token>, LexerError> {
        let c = loop {
            match self.current() {
                Some(c) if c.is_ascii_whitespace() => {
                    self.advance();
                }
                Some(c) => break c,
                None => return Ok(None),
            }
        };

        let mut span = self.span();
        let kind = match c {
            '#' if matches!(self.peek(1), Some('|')) => {
                self.advance(); // '#'
                self.advance(); // '|#'

                let mut comment = String::new();
                while let Some(c) = self.advance() {
                    match c {
                        '|' if matches!(self.peek(0), Some('#')) => {
                            self.advance(); // '|'
                            self.advance(); // '#'
                            break;
                        }
                        c => {
                            comment.push(c);
                        }
                    }
                }

                TokenKind::BlockComment(comment.trim().into())
            }
            ';' => {
                let mut comment = String::new();
                while let Some(c) = self.advance() {
                    match c {
                        ';' => continue,
                        '\n' => break,
                        c => {
                            comment.push(c);
                        }
                    }
                }

                TokenKind::LineComment(comment.trim().into())
            }

            '(' => {
                self.advance();
                TokenKind::OpenParen
            }
            ')' => {
                self.advance();
                TokenKind::CloseParen
            }
            '{' => {
                self.advance();
                TokenKind::OpenBrace
            }
            '}' => {
                self.advance();
                TokenKind::CloseBrace
            }
            '[' => {
                self.advance();
                TokenKind::OpenBracket
            }
            ']' => {
                self.advance();
                TokenKind::CloseBracket
            }
            '#' if matches!(self.peek(1), Some('{')) => {
                self.advance(); // '#'
                self.advance(); // '{'
                TokenKind::OpenHashBrace
            }

            '\'' => {
                self.advance(); // '\''

                let c = match self.advance() {
                    Some('\\') => match self.advance() {
                        Some(c @ ('"' | '\\')) => c,
                        Some('n') => '\n',
                        Some('e') => '\x1b',
                        Some(c) => {
                            return Err(LexerError::new(
                                LexerErrorKind::InvalidEscape(c),
                                span.join(&self.span()),
                            ));
                        }
                        None => {
                            return Err(LexerError::new(
                                LexerErrorKind::UnclosedChar,
                                span.join(&self.span()),
                            ));
                        }
                    },
                    Some(c) => c,
                    None => {
                        return Err(LexerError::new(
                            LexerErrorKind::UnclosedChar,
                            span.join(&self.span()),
                        ))
                    }
                };

                if self.advance() != Some('\'') {
                    self.read_word();
                    return Err(LexerError::new(
                        LexerErrorKind::InvalidChar,
                        span.join(&self.span()),
                    ));
                }

                TokenKind::Char(c)
            }
            '0'..='9' => TokenKind::Number(self.parse_or(LexerErrorKind::InvalidNumber)?),
            '+' | '-' if matches!(self.peek(1), Some('0'..='9')) => {
                TokenKind::Number(self.parse_or(LexerErrorKind::InvalidNumber)?)
            }
            '"' => {
                self.advance(); // '"'

                let quote_span = span.clone().join(&self.span());
                let mut string = String::new();

                loop {
                    let ch_span = self.span();
                    string.push(match self.advance() {
                        Some('"') => break,
                        Some('\\') => match self.advance() {
                            Some(c @ ('"' | '\\')) => c,
                            Some('n') => '\n',
                            Some('e') => '\x1b',
                            Some(c) => {
                                return Err(LexerError::new(
                                    LexerErrorKind::InvalidEscape(c),
                                    ch_span.join(&self.span()),
                                ))
                            }
                            None => {
                                return Err(LexerError::new(
                                    LexerErrorKind::UnclosedString,
                                    quote_span,
                                ))
                            }
                        },
                        Some(c) => c,
                        None => {
                            return Err(LexerError::new(LexerErrorKind::UnclosedString, quote_span))
                        }
                    });
                }

                if self.current().is_some_and(|c| !c.is_separator()) {
                    self.read_word();
                    return Err(LexerError::new(
                        LexerErrorKind::InvalidString,
                        span.join(&self.span()),
                    ));
                }

                TokenKind::String(string)
            }
            ':' => {
                self.advance();
                TokenKind::Keyword(Symbol(self.read_word()))
            }
            _ => {
                let word = self.read_word();
                match word.as_str() {
                    "true" => TokenKind::Bool(true),
                    "false" => TokenKind::Bool(false),
                    "nil" => TokenKind::Nil,
                    _ => TokenKind::Symbol(Symbol::from(word)),
                }
            }
        };

        span.extend(&self.span());

        Ok(Some(Token::new(kind, span)))
    }
}

impl Iterator for Lexer<'_> {
    type Item = Result<Token, LexerError>;

    fn next(&mut self) -> Option<Self::Item> {
        self.read().transpose()
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    macro_rules! test {
        ( $name:ident: $input:literal, $tokens:expr ) => {
            #[test]
            fn $name() {
                let mut lexer = Lexer::new($input);

                for token in $tokens {
                    let x = lexer.next().map(|r| match r {
                        Ok(t) => Ok(t.kind),
                        Err(e) => Err(e.kind),
                    });

                    assert_eq!(x, Some(token));
                }

                assert_eq!(lexer.next(), None);
            }
        };
    }

    test!(block_comment: "#| foo\nbar |#", [
        Ok(TokenKind::BlockComment("foo\nbar".into()))
    ]);

    test!(line_comment: "; foo\n;; bar baz qux", [
        Ok(TokenKind::LineComment("foo".into())),
        Ok(TokenKind::LineComment("bar baz qux".into())),
    ]);

    test!(hashset: "{{} true false}", [
        Ok(TokenKind::OpenBrace),
        Ok(TokenKind::OpenBrace),
        Ok(TokenKind::CloseBrace),
        Ok(TokenKind::Bool(true)),
        Ok(TokenKind::Bool(false)),
        Ok(TokenKind::CloseBrace),
    ]);

    test!(hashmap: "(foo #{:bar 0 :baz 1})", [
        Ok(TokenKind::OpenParen),
        Ok(TokenKind::Symbol(Symbol::from("foo"))),
        Ok(TokenKind::OpenHashBrace),
        Ok(TokenKind::Keyword(Symbol::from("bar"))),
        Ok(TokenKind::Number(0.0)),
        Ok(TokenKind::Keyword(Symbol::from("baz"))),
        Ok(TokenKind::Number(1.0)),
        Ok(TokenKind::CloseBrace),
        Ok(TokenKind::CloseParen),
    ]);

    test!(char_literal: "'x' '\n'", [
        Ok(TokenKind::Char('x')),
        Ok(TokenKind::Char('\n')),
    ]);

    test!(lex: "(+ 14 25.5 333 (* 2 5))", [
        Ok(TokenKind::OpenParen),
        Ok(TokenKind::Symbol(Symbol::from("+"))),
        Ok(TokenKind::Number(14.0)),
        Ok(TokenKind::Number(25.5)),
        Ok(TokenKind::Number(333.0)),
        Ok(TokenKind::OpenParen),
        Ok(TokenKind::Symbol(Symbol::from("*"))),
        Ok(TokenKind::Number(2.0)),
        Ok(TokenKind::Number(5.0)),
        Ok(TokenKind::CloseParen),
        Ok(TokenKind::CloseParen),
    ]);

    test!(newline: "(+ 14 25.5 333\n(* 2 5 5.x))", [
        Ok(TokenKind::OpenParen),
        Ok(TokenKind::Symbol(Symbol::from("+"))),
        Ok(TokenKind::Number(14.0)),
        Ok(TokenKind::Number(25.5)),
        Ok(TokenKind::Number(333.0)),
        Ok(TokenKind::OpenParen),
        Ok(TokenKind::Symbol(Symbol::from("*"))),
        Ok(TokenKind::Number(2.0)),
        Ok(TokenKind::Number(5.0)),
        Err(LexerErrorKind::InvalidNumber("5.x".into())),
        Ok(TokenKind::CloseParen),
        Ok(TokenKind::CloseParen),
    ]);

    test!(negative_minus: "(- 1 -2 3)", [
        Ok(TokenKind::OpenParen),
        Ok(TokenKind::Symbol(Symbol::from("-"))),
        Ok(TokenKind::Number(1.0)),
        Ok(TokenKind::Number(-2.0)),
        Ok(TokenKind::Number(3.0)),
        Ok(TokenKind::CloseParen),
    ]);

    test!(error_parse_numbers: "2 55 3.144 0.0001 1.1.1", [
        Ok(TokenKind::Number(2.0)),
        Ok(TokenKind::Number(55.0)),
        Ok(TokenKind::Number(3.144)),
        Ok(TokenKind::Number(0.0001)),
        Err(LexerErrorKind::InvalidNumber("1.1.1".into())),
    ]);

    test!(error_unclosed_string: "\"hiii", [
        Err(LexerErrorKind::UnclosedString),
    ]);

    test!(error_invalid_string: "\"hiii\"222", [
        Err(LexerErrorKind::InvalidString),
    ]);
}
